﻿Imports EventBindings.Extensions
Imports System.ComponentModel.Design
Imports System.Reflection
Public Class MethodBindingProvider
    Implements IEventHandlerProvider
    Implements IEventMediatorFactory

#Region "Fields"
    Private _Factory As New MethodMediatorFactory
#End Region

#Region "Methods"
    Private Function GetSupportedMethods(ByVal source As Object) As IEnumerable(Of MethodInfo)
        Dim SupportedMethods As New List(Of MethodInfo)
        Dim Names As New Specialized.StringCollection
        Dim Methods = source.GetType.GetMethods
        For Each method In Methods
            If Names.Contains(method.Name) Then Continue For
            If method.IsAbstract Then Continue For
            If method.IsConstructor Then Continue For
            If method.IsGenericMethodDefinition Then Continue For
            If method.ContainsGenericParameters Then Continue For
            If Not Me.AllowSpecialNamedMethod AndAlso method.IsSpecialName Then Continue For
            If Not Me.IsSupportedMethod(method) Then Continue For
            SupportedMethods.Add(method)
            Names.Add(method.Name)
        Next
        Return SupportedMethods
    End Function
    Private Function IsSupportedMethod(ByVal mi As MethodInfo) As Boolean
        Dim Parameters = mi.GetParameters
        If Parameters.IsNullOrEmpty Then Return True
        Return Parameters.Count <= 2
    End Function
    Private Function GetMethod(ByVal source As Object, ByVal name As String) As MethodInfo
        If source Is Nothing Then Return Nothing
        Dim SupportedMethods = Me.GetSupportedMethods(source)
        Return SupportedMethods.Take(Function(M)
                                         Return String.Equals(M.Name, name, StringComparison.InvariantCultureIgnoreCase)
                                     End Function)
    End Function
#End Region

#Region "Properties"
    Public Property AllowSpecialNamedMethod As Boolean = True
    Public Property Synchronizer As ISynchronizeInvoke
        Get
            Return Me._Factory.Synchronizer
        End Get
        Set(ByVal value As ISynchronizeInvoke)
            Me._Factory.Synchronizer = value
        End Set
    End Property
#End Region

#Region "IEventHandlerProvider Implementation"
    Public Function CanResolveHandler(ByVal eventHandlerID As Object, ByVal eventHandlerSource As Object) As Boolean Implements IEventHandlerProvider.CanResolveHandler
        If eventHandlerSource Is Nothing Then Return False
        If Not TypeOf eventHandlerID Is String Then Return False
        Dim SupportedMethods = Me.GetSupportedMethods(eventHandlerSource)
        Dim Name = DirectCast(eventHandlerID, String)
        Return SupportedMethods.Contains(Function(M)
                                             Return String.Equals(M.Name, Name, StringComparison.InvariantCultureIgnoreCase)
                                         End Function)
    End Function
    Public Function GetEventHandlerIDs(ByVal eventHandlerSource As Object) As IEnumerable Implements IEventHandlerProvider.GetEventHandlerIDs
        If eventHandlerSource Is Nothing Then Return Nothing
        Dim SupportedMethods = Me.GetSupportedMethods(eventHandlerSource)
        Dim Names As New Specialized.StringCollection
        For Each method In SupportedMethods
            Names.Add(method.Name)
        Next
        Return Names
    End Function

    Public Function GetEventHandlerSources() As IEnumerable Implements IEventHandlerProvider.GetEventHandlerSources
        If Me.Site Is Nothing Then Return Nothing
        Dim Host = Me.Site.GetService(Of IDesignerHost)()
        If Host Is Nothing Then Return Nothing
        Return Host.Container.Components
    End Function

    Public Function ResolveHandler(ByVal eventHandlerID As Object, ByVal eventHandlerSource As Object) As Object Implements IEventHandlerProvider.ResolveHandler
        If Not TypeOf eventHandlerID Is String Then Return False
        Return Me.GetMethod(eventHandlerSource, DirectCast(eventHandlerID, String))
    End Function
    Public Function ResolveHandlerType(ByVal eventHandlerID As Object, ByVal eventHandlerSource As Object) As Type Implements IEventHandlerProvider.ResolveHandlerType
        If Not TypeOf eventHandlerID Is String Then Return Nothing
        Dim Handler = Me.GetMethod(eventHandlerSource, DirectCast(eventHandlerID, String))
        If Handler Is Nothing Then Return Nothing
        Return Handler.GetType
    End Function
#End Region

#Region "IEventMediatorFactory"
    Public Function SupportMediationFor(ByVal [event] As EventData, ByVal handling As HandlingData, ByVal mediationMode As MediationMode) As Boolean Implements IEventMediatorFactory.SupportMediationFor
        Return Me._Factory.SupportMediationFor([event], handling, mediationMode)
    End Function
    Public Function CreateMediator(ByVal [event] As EventData, ByVal handling As HandlingData, ByVal mediationMode As MediationMode) As IEventMediator Implements IEventMediatorFactory.CreateMediator
        Return Me._Factory.CreateMediator([event], handling, mediationMode)
    End Function
#End Region

End Class
