﻿Imports System.Reflection
Imports System.Reflection.Emit
Imports EventBindings.Extensions

Public Class MethodMediator(Of TEventArgs As EventArgs)
    Implements IMethodMediator

#Region "Fields"
    Private _Asynchronous As Boolean = False
    Private _Synchronizer As ISynchronizeInvoke
    Private _Method As MethodInfo
    Private _MethodSource As Object
    Private _Event As EventData
    Private _Handler As [Delegate] ' Action(Of Object, TEventArgs)
#End Region

#Region "Properties"
    Protected ReadOnly Property EventType As Type
        Get
            Return Me._Event.EventDescriptor.EventType
        End Get
    End Property
#End Region

#Region "Methods"
    Public Overloads Sub Initialize(ByVal [event] As EventData) Implements IEventMediator.Initialize
        If Not [event].IsValidEvent Then Throw New InvalidEventException([event])
        Me._Event = [event]
    End Sub
    Public Overloads Sub Initialize(ByVal method As MethodInfo, ByVal methodSource As Object) Implements IMethodMediator.Initialize
        If method Is Nothing Then Throw New ArgumentNullException("method")
        If methodSource Is Nothing Then Throw New ArgumentNullException("methodSource")
        Me._Method = method
        Me._MethodSource = methodSource
    End Sub
    Public Overloads Sub Initialize(ByVal mode As MediationMode, Optional ByVal synchronizer As ISynchronizeInvoke = Nothing) Implements IMethodMediator.Initialize
        Select Case mode
            Case MediationMode.Synchronized
                If synchronizer Is Nothing Then Throw New ArgumentNullException("synchronizer")
                Me._Synchronizer = synchronizer
            Case MediationMode.Asynchronous
                Me._Asynchronous = True
            Case Else
                Me._Asynchronous = False
        End Select
    End Sub
    Public Sub [AddHandler]() Implements IEventMediator.AddHandler
        If Me._Handler IsNot Nothing Then Exit Sub
        Me._Handler = Me.CreateHandler
        Me._Event.AddEventHandler(Me._Handler)
    End Sub
    Public Sub [RemoveHandler]() Implements IEventMediator.RemoveHandler
        If Me._Handler Is Nothing Then Exit Sub
        Me._Event.RemoveEventHandler(Me._Handler)
        Me._Handler = Nothing
    End Sub

    Protected Overridable Function CreateHandler() As [Delegate]
        If Me.IsSynchronized Then
            Return New EventHandler(Of TEventArgs)(AddressOf Me.OnSynchronizedEvent)
        ElseIf Me.IsAsynchronous Then
            Return New EventHandler(Of TEventArgs)(AddressOf Me.OnAsynchronousEvent)
        Else
            Return New EventHandler(Of TEventArgs)(AddressOf Me.OnSynchronousEvent)
        End If
    End Function

    Private Function GetMethodArguments(ByVal sender As Object, ByVal e As TEventArgs) As Object()
        If Not Me._Method.HasParameters Then Return Nothing
        Dim Parameters = Me._Method.GetParameters
        Select Case Parameters.Count
            Case 1
                Dim Parameter = Parameters(0)
                Dim Argument = Parameter.GetArgumentValue(sender, e)
                Return New Object() {Argument}
            Case 2
                Dim Argument1 = Parameters(0).GetArgumentValue(sender, e)
                Dim Argument2 = Parameters(1).GetArgumentValue(sender, e)
                Return New Object() {Argument1, Argument2}
            Case Else
                Throw New NotSupportedException
        End Select
    End Function
    Protected Function IsSynchronized() As Boolean
        Return Me._Synchronizer IsNot Nothing
    End Function
    Protected Function IsAsynchronous() As Boolean
        Return Me._Asynchronous
    End Function

#End Region

#Region "Event Handlers"
    Protected Sub OnSynchronousEvent(ByVal sender As Object, ByVal e As TEventArgs)
        Dim Arguments = Me.GetMethodArguments(sender, e)
        Me._Method.Invoke(Me._MethodSource, Arguments)
    End Sub
    Protected Sub OnAsynchronousEvent(ByVal sender As Object, ByVal e As TEventArgs)
        Static Invoker As New EventHandler(Of TEventArgs)(AddressOf Me.OnSynchronousEvent)
        Invoker.BeginInvoke(sender, e, Nothing, Nothing)
    End Sub
    Protected Sub OnSynchronizedEvent(ByVal sender As Object, ByVal e As TEventArgs)
        If Me._Synchronizer.InvokeRequired Then
            Static Invoker As New EventHandler(Of TEventArgs)(AddressOf Me.OnSynchronousEvent)
            Me._Synchronizer.Invoke(Invoker, New Object() {sender, e})
        Else
            Me.OnSynchronousEvent(sender, e)
        End If
    End Sub
#End Region
End Class
